@@ -39,56 +39,74 @@ def _seq_but_not_str(obj: object) -> bool:
 
 ## currently the files spec is missing operationId's for some of its operations.
 RESOURCES = {
+    'local': ['actors','authenticator', 'meta', 'files', 'sk', 'streams', 'systems', 'tenants','tokens'],
     'tapipy':{
-        'actors': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-actors.yml',
-        'authenticator': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-authenticator.yml',
-        'meta': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-meta.yml',
-        'files': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-files.yml',
-        'sk': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-sk.yml',
-        'streams': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-streams.yml',
-        'systems': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-systems.yml',
-        'tenants': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-tenants.yml',
-        'tokens': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-tokens.yml'
+        'actors': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-actors.yml',
+        'authenticator': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-authenticator.yml',
+        'meta': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-meta.yml',
+        'files': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-files.yml',
+        'sk': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-sk.yml',
+        'streams': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-streams.yml',
+        'systems': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-systems.yml',
+        'tenants': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-tenants.yml',
+        'tokens': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-tokens.yml'
     },
-    'prod': {
-        'actors': 'https://raw.githubusercontent.com/TACC/abaco/dev-v3/docs/specs/openapi_v3.yml',               
-        'authenticator': 'https://raw.githubusercontent.com/tapis-project/authenticator/prod/service/resources/openapi_v3.yml',
-        'meta': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/prod/meta-client/src/main/resources/metav3-openapi.yaml',
-        'files': 'https://raw.githubusercontent.com/tapis-project/tapis-files/prod/api/src/main/resources/openapi.yaml',
-        'sk': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/prod/security-client/src/main/resources/SKAuthorizationAPI.yaml',
-        'streams': 'https://raw.githubusercontent.com/tapis-project/streams-api/prod/service/resources/openapi_v3.yml',
-        'systems': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/prod/systems-client/SystemsAPI.yaml',
-        'tenants': 'https://raw.githubusercontent.com/tapis-project/tenants-api/prod/service/resources/openapi_v3.yml',
-        'tokens': 'https://raw.githubusercontent.com/tapis-project/tokens-api/prod/service/resources/openapi_v3.yml'
+    'master': {
+        'actors': 'https://raw.githubusercontent.com/TACC/abaco/master/docs/specs/openapi_v3.yml',               
+        'authenticator': 'https://raw.githubusercontent.com/tapis-project/authenticator/dev/service/resources/openapi_v3.yml',
+        'meta': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/master/meta-client/src/main/resources/metav3-openapi.yaml',
+        'files': 'https://raw.githubusercontent.com/tapis-project/tapis-files/master/api/src/main/resources/openapi.yaml',
+        'sk': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/master/security-client/src/main/resources/SKAuthorizationAPI.yaml',
+        'streams': 'https://raw.githubusercontent.com/tapis-project/streams-api/dev/service/resources/openapi_v3.yml',
+        'systems': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/master/systems-client/SystemsAPI.yaml',
+        'tenants': 'https://raw.githubusercontent.com/tapis-project/tenants-api/master/service/resources/openapi_v3.yml',
+        'tokens': 'https://raw.githubusercontent.com/tapis-project/tokens-api/master/service/resources/openapi_v3.yml'
     },
     'dev': {
-        'actors': 'https://raw.githubusercontent.com/TACC/abaco/dev-v3/docs/specs/openapi_v3.yml',
+        'actors': 'https://raw.githubusercontent.com/TACC/abaco/master/docs/specs/openapi_v3.yml',
         'authenticator': 'https://raw.githubusercontent.com/tapis-project/authenticator/dev/service/resources/openapi_v3.yml',
-        'meta': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/dev/meta-client/src/main/resources/metav3-openapi.yaml',
-        'files': 'https://raw.githubusercontent.com/tapis-project/tapis-files/dev/api/src/main/resources/openapi.yaml',
+        'meta': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/master/meta-client/src/main/resources/metav3-openapi.yaml',
+        'files': 'https://raw.githubusercontent.com/tapis-project/tapis-files/master/api/src/main/resources/openapi.yaml',
         'sk': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/dev/security-client/src/main/resources/SKAuthorizationAPI.yaml',
         'streams': 'https://raw.githubusercontent.com/tapis-project/streams-api/dev/service/resources/openapi_v3.yml',
-        'systems': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/dev/systems-client/SystemsAPI.yaml',
-        'tenants': 'https://raw.githubusercontent.com/tapis-project/tenants-api/dev/service/resources/openapi_v3.yml',
-        'tokens': 'https://raw.githubusercontent.com/tapis-project/tokens-api/dev/service/resources/openapi_v3.yml'
+        'systems': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/master/systems-client/SystemsAPI.yaml',
+        'tenants': 'https://raw.githubusercontent.com/tapis-project/tenants-api/master/service/resources/openapi_v3.yml',
+        'tokens': 'https://raw.githubusercontent.com/tapis-project/tokens-api/master/service/resources/openapi_v3.yml'
     }
 }
 
 
+def get_specs_from_local_resource_defs(resources):
+    specs = {}
+    resources_dir = os.path.join(os.path.dirname(__file__), 'resources')
+    for r in resources:
+        try:
+            # for now, hardcode the paths; we could look these up based on a canonical URL once that is
+            # established.
+            spec_path = f'{resources_dir}/openapi_v3-{r}.yml'
+            spec_dict = yaml.load(open(spec_path, 'r'), Loader=yaml.FullLoader)
+            specs[r] = create_spec(spec_dict)
+        except:
+            pass
+    return specs
+
 
 def _get_specs(resources: Resources, spec_dir: str = None, download_latest_specs: bool = False) -> Specs:
     """
     Gets specs requested in resources.
     Will download any spec that is not already downloaded.
     """
+    # if no resources passed, return the specs corresponding to the resource folder --
+    if type(resources) == list:
+        return get_specs_from_local_resource_defs(resources)
     spec_dir = get_spec_dir(spec_dir)
     # Download and save specs if neccessary
-    download_and_pickle_spec_dicts(resources, spec_dir=spec_dir, download_latest_specs=download_latest_specs)
+    download_and_pickle_spec_dicts(resources.values(), spec_dir=spec_dir, download_latest_specs=download_latest_specs)
     # Load, unpickle, and create specs
     specs = unpickle_and_create_specs(resources, spec_dir=spec_dir)
     return specs
 
-def download_and_pickle_spec_dicts(resources: Resources, spec_dir: str, download_latest_specs: bool) -> None:
+def download_and_pickle_spec_dicts(url_list: list, spec_dir: str, download_latest_specs: bool) -> None:
     """
     Function that calls threads to download and pickle specs.
     Cuts wait time for 9 specs from 4s to 0.7s
@@ -96,15 +114,13 @@ def download_and_pickle_spec_dicts(resources: Resources, spec_dir: str, download
     # Get list of specs and check which we need to download
     # We download if the file name requested does not exist
     urls_to_download = []
-    for resource_name, url in resources.items():
-        if "local:" in url:
-            continue 
+    for url in set(url_list):
         _, full_spec_name, spec_path = get_file_info_from_url(url, spec_dir)
         if download_latest_specs:
-            urls_to_download.append([resource_name, url, spec_path])
+            urls_to_download.append([url, spec_path])
         else:
             if not full_spec_name in os.listdir(spec_dir):
-                urls_to_download.append([resource_name, url, spec_path])
+                urls_to_download.append([url, spec_path])
     
     # Set off some parallel processes cause it's quick.
     POOL_SIZE = os.environ.get('POOL_SIZE', 16)
@@ -119,7 +135,7 @@ def _thread_download_spec_dict(resource_info: ResourceInfo) -> None:
     spec dicts. Gets spec dict, if it's valid, stores it at 'spec_path',
     else it does nothing.
     """
-    resource_name, resource_url, spec_path = resource_info
+    resource_url, spec_path = resource_info
     # Attempt to get spec from url
     response = requests.get(resource_url)
     if response.status_code == 200:
@@ -127,8 +143,8 @@ def _thread_download_spec_dict(resource_info: ResourceInfo) -> None:
             # Loads yaml into Python dictionary
             spec_dict = yaml.load(response.content, Loader=yaml.FullLoader)
         except Exception as e:
-            print(f'Got exception when attempting to load yaml for '
-                  f'"{resource_name}" resource; exception: {e}')
+            print(f'Got exception when attempting to load yaml for'
+                  f'"{spec_path}" resource; exception: {e}')
             return
         try:
             # Attempts to create spec from dict to ensure the spec is valid
@@ -137,19 +153,19 @@ def _thread_download_spec_dict(resource_info: ResourceInfo) -> None:
             test_spec_dict = copy.deepcopy(spec_dict)
             create_spec(test_spec_dict)
         except Exception as e:
-            print(f'Got exception when test creating spec for "{resource_name}" '
+            print(f'Got exception when test creating spec for "{spec_path}" '
                   f'resource; Spec probably not verifying; exception: {e}')
             return
         try:
             # Pickles and saves the spec dict to the spec_path atomically
             with atomic_write(f'{spec_path}', overwrite=True, mode='wb') as spec_file:
-                pickle.dump(spec_dict, spec_file, protocol=4)
+                pickle.dump(spec_dict, spec_file, protocol=pickle.HIGHEST_PROTOCOL)
         except Exception as e:
-            print(f'Got exception when attempting to pickle spec_dict for '
-                  f'"{resource_name}" resource; exception: {e}')
+            print(f'Got exception when attempting to pickle spec_dict for'
+                  f'"{spec_path}" resource; exception: {e}')
             return
     else:
-        raise KeyError(f'Error getting "{resource_name}" resource. URL: "{resource_url}".'
+        raise KeyError(f'Error getting "{spec_path}" resource. URL: "{resource_url}".'
                        f'Did not get 200, got the following back:\n{response.text}')
     
 def unpickle_and_create_specs(resources: Resources, spec_dir: str) -> Specs:
@@ -160,16 +176,6 @@ def unpickle_and_create_specs(resources: Resources, spec_dir: str) -> Specs:
     specs = {}
     # Get resource path to point the unpickling at.
     for resource_name, url in resources.items():
-        if "local:" in url:
-            try:
-                spec_path = url.replace('local:', '')
-                with open(spec_path, 'rb') as spec_file:
-                    spec_dict = yaml.load(spec_file, Loader=yaml.FullLoader)
-                specs.update({resource_name: create_spec(spec_dict)})
-            except Exception as e:
-                print(f'Error reading local "{resource_name}" resource. '
-                      f'Ensure path is absolute. e:{e}')
-            continue
         _, _, spec_path = get_file_info_from_url(url, spec_dir)
         try:
             # Unpickle and create_spec
@@ -177,24 +183,8 @@ def unpickle_and_create_specs(resources: Resources, spec_dir: str) -> Specs:
                 spec_dict = pickle.load(spec_file)
             specs.update({resource_name: create_spec(spec_dict)})
         except Exception as e:
-            print(f'Got exception trying to unpickle and create spec for '
-                  f'spec_path: "{resource_name}"; exception: {e}')
-            # Using resource name to look if the resource exists in the
-            # tapipy prod resources. If it is found, we fall back and use that!
-            # If it isn't we just skip it.
-            print(f'Falling back to prod spec for "{resource_name}" resource')
-            if resource_name in RESOURCES['tapipy']:
-                url = RESOURCES['tapipy'][resource_name]
-                _, _, spec_path = get_file_info_from_url(url, spec_dir)
-                try:
-                    # Unpickle and create_spec
-                    with open(spec_path, 'rb') as spec_file:
-                        spec_dict = pickle.load(spec_file)
-                    specs.update({resource_name: create_spec(spec_dict)})
-                except Exception as e:
-                    print('Error opening tapipy prod spec. This is bad.')
-            else:
-                print(f'No "{resource_name}" spec was found to fallback on')
+            print(f'Got exception trying to unpickle and create spec for'
+                  f'spec_path: "{spec_path}"; exception: {e}')
     return specs
 
 def update_spec_cache(resources: Resources = None, spec_dir: str = None) -> None:
@@ -263,6 +253,40 @@ def get_basic_auth_header(username: str, password: str) -> str:
     return 'Basic {}'.format(b64encode(user_pass).decode())
 
 
+class Tenants(object):
+    """
+    Basic tenancy manager for Tapipy clients.
+    """
+
+    def __init__(self, tapi_client):
+        self.tapi_client = tapi_client
+        self.tenants = {}
+
+    def reload_tenants(self):
+        try:
+            sites = self.tapi_client.tenants.list_sites()
+            tenants = self.tapi_client.tenants.list_tenants()
+        except Exception as e:
+            raise errors.BaseTapyException(f"Unable to retrieve sites and tenants from the Tenants API. e: {e}")
+        for t in tenants:
+            for s in sites:
+                if s.site_id == t.site_id:
+                    t.site = s
+
+        self.tenants = {tn.tenant_id: tn for tn in tenants}
+
+    def get_tenant_config(self, tenant_id):
+        try:
+            return self.tenants[tenant_id]
+        except KeyError:
+            # update the cache and try again, but let the KeyError bubble up this time:
+            self.reload_tenants()
+            try:
+                return self.tenants[tenant_id]
+            except KeyError:
+                raise errors.BaseTapyException("Tenant not found.")
+
+
 class Tapis(object):
     """
     A client for the Tapis API.
@@ -286,7 +310,9 @@ class Tapis(object):
                  resource_set: str = 'tapipy',
                  custom_spec_dict: Resources = None,
                  download_latest_specs: bool = False,
-                 spec_dir: str = None
+                 spec_dir: str = None,
+                 tenants: Tenants = None,
+                 is_tapis_service: bool = False
                  ):
         # the base_url for the server this Tapis client should interact with
         self.base_url = base_url
@@ -364,7 +390,7 @@ class Tapis(object):
         self.spec_dir = spec_dir
 
         # Uses module instantiated RESOURCE_SPECS if there are no changes to the specs. 
-        if self.custom_spec_dict or self.spec_dir or self.download_latest_specs or not self.resource_set == 'tapipy':
+        if self.custom_spec_dict or self.spec_dir or self.download_latest_specs or not self.resource_set == 'master':
             resource_specs = _get_specs(RESOURCES[resource_set], spec_dir=self.spec_dir, download_latest_specs=self.download_latest_specs)
         else:
             resource_specs = RESOURCE_SPECS
@@ -374,12 +400,18 @@ class Tapis(object):
             # each API is a top-level attribute on the DynaTapy object, a Resource object constructed as follows:
             setattr(self, resource_name, Resource(resource_name, spec.paths, self))
 
+        # whether or not this tapis client is running inside a Tapis service.
+        self.is_tapis_service = is_tapis_service
+
         # we lazy-load the tenant_cache to prevent making a call to the Tenants API when not needed.
-        self.tenant_cache = {}
+        if tenants:
+            self.tenant_cache = tenants
+        else:
+            self.tenant_cache = Tenants(self)
         # if the user passed just base_url, try to get the list of tenants and derive the tenant_id from it.
-        if base_url and not tenant_id:
-            self.update_tenant_cache()
-            for tid, t in self.tenant_cache.items():
+        if not self.is_tapis_service and base_url and not tenant_id:
+            self.tenant_cache.reload_tenants()
+            for tid, t in self.tenant_cache.tenants.items():
                 if t.base_url == base_url:
                     self.tenant_id = t.tenant_id
 
@@ -390,6 +422,7 @@ class Tapis(object):
                 self.x_tenant_id = self.tenant_id
                 self.x_username = self.username
 
+
     def update_spec_cache(self):
         """
         Updates the spec cache by using the Tapis object's settings.
@@ -406,30 +439,6 @@ class Tapis(object):
                     raise KeyError(f"Custom spec should be a dict of key: str and val: str, got {spec_name} and {spec_val}.")
         update_spec_cache(RESOURCES[self.resource_set], self.spec_dir)
 
-    def update_tenant_cache(self):
-        """
-        Update the local cache of tenant objects.
-        :return:
-        """
-        self.tenant_cache = {tn.tenant_id:tn for tn in self.tenants.list_tenants()}
-
-    def get_tenant_config(self, tenant_id):
-        """
-        Retrieve the configuration for a tenant. Uses the cache
-        :param tenant_id:
-        :return:
-        """
-        try:
-            return self.tenant_cache[tenant_id]
-        except KeyError:
-            # update the cache and try again, but let the KeyError bubble up this time:
-            self.update_tenant_cache()
-            try:
-                return self.tenant_cache[tenant_id]
-            except KeyError:
-                raise errors.BaseTapyException("Tenant not found.")
-
-
     def get_tokens(self, **kwargs):
         """
         Convenience wrapper to get either service tokens (tokengen Tokens API) or user tokens (Authenticator/OAuth2 API)
@@ -477,17 +486,26 @@ class Tapis(object):
 
     def get_service_tokens(self, **kwargs):
         """
-        Calls the Tapis Tokens API (tokengen) to get access and refresh tokens for a service and set them on the client.
+        Calls the Tapis Tokens API to get access and refresh tokens for a service and set them on the client.
         :return:
         """
         if not 'username' in kwargs:
             username = self.username
         else:
             username = kwargs['username']
-        if not 'tenant_id' in kwargs:
-            tenant_id = self.tenant_id
+        # tapis services manage a set ot tokens, one for each of the tenants for which we need to interact with.
+        self.service_tokens = {}
+        # if the caller passed a tenant_id explicitly, we just use that
+        if 'tenant_id' in kwargs:
+            self.service_tokens = {kwargs['tenant_id']: {}}
+        # otherwise, we compute all the tenant's that this service could need to interact with.
         else:
-            tenant_id = kwargs['tenant_id']
+            try:
+                self.service_tokens = {t: {} for t in self.tenant_cache.get_site_master_tenants_for_service()}
+            except AttributeError:
+                raise errors.BaseTapyException("Unable to retrieve target tenants for a service. "
+                                               "Are you really a Tapis service? "
+                                               "Did you pass in your Tenants instance?")
         if not 'access_token_ttl' in kwargs:
             # default to a 24 hour access token -
             access_token_ttl = 86400
@@ -498,19 +516,23 @@ class Tapis(object):
             refresh_token_ttl = 3153600000
         else:
             refresh_token_ttl = kwargs['refresh_token_ttl']
-        if 'service_password' in kwargs:
-            service_password = kwargs['service_password']
-        elif self.service_password:
-            service_password = self.service_password
-
-        tokens = self.tokens.create_token(token_username=username,
-                                          token_tenant_id=tenant_id,
-                                          account_type=self.account_type,
-                                          access_token_ttl=access_token_ttl,
-                                          generate_refresh_token=True,
-                                          refresh_token_ttl=refresh_token_ttl)
-        self.set_access_token(tokens.access_token)
-        self.set_refresh_token(tokens.refresh_token)
+        for tenant_id in self.service_tokens.keys():
+            try:
+                target_site_id = self.tenant_cache.get_tenant_config(tenant_id=tenant_id).site_id
+            except Exception as e:
+                raise errors.BaseTapyException(f"Got exception computing target site id; e:{e}")
+            try:
+                tokens = self.tokens.create_token(token_username=username,
+                                                  token_tenant_id=self.tenant_id,
+                                                  account_type=self.account_type,
+                                                  access_token_ttl=access_token_ttl,
+                                                  generate_refresh_token=True,
+                                                  refresh_token_ttl=refresh_token_ttl,
+                                                  target_site_id=target_site_id)
+            except Exception as e:
+                raise errors.BaseTapyException(f"Could not generate tokens for tenant: {tenant_id}; exception: {e}")
+            self.service_tokens[tenant_id] = {'access_token': tokens.access_token,
+                                              'refresh_token': tokens.refresh_token}
 
     def validate_token(self, token):
         """
@@ -534,7 +556,7 @@ class Tapis(object):
                 "Unable to process Tapis token; could not parse the tenant_id. It is possible "
                 "the token is in a format no longer supported by the platform.")
         try:
-            public_key_str = self.get_tenant_config(tenant_id=token_tenant_id).public_key
+            public_key_str = self.tenant_cache.get_tenant_config(tenant_id=token_tenant_id).public_key
         except errors.BaseTapyException:
             raise errors.TokenInvalidError("Unable to process Tapis token; unexpected tenant_id.")
         except KeyError:
@@ -565,14 +587,14 @@ class Tapis(object):
         except:
             pass
 
-    def refresh_tokens(self):
+    def refresh_tokens(self ,tenant_id=None):
         """
         Use the refresh token on this client to get a new access and refresh token pair.
         """
         if not self.refresh_token:
             raise errors.TapyClientConfigurationError(msg="No refresh token found.")
         if self.account_type == 'service':
-            return self.refresh_service_tokens()
+            return self.refresh_service_tokens(tenant_id)
         else:
             return self.refresh_user_tokens()
 
@@ -591,13 +613,14 @@ class Tapis(object):
         self.set_access_token(tokens.access_token)
         self.set_refresh_token(tokens.refresh_token)
 
-    def refresh_service_tokens(self):
+    def refresh_service_tokens(self, tenant_id):
         """
         Use the refresh token operation for tokens of type "service".
         """
-        tokens = self.tokens.refresh_token(refresh_token=self.refresh_token.refresh_token)
-        self.set_access_token(tokens.access_token)
-        self.set_refresh_token(tokens.refresh_token)
+        refresh_token = self.service_tokens[tenant_id]['refresh_token'].refresh_token
+        tokens = self.tokens.refresh_token(refresh_token=refresh_token)
+        self.service_tokens[tenant_id]['access_token'] = tokens.refresh_token
+        self.service_tokens[tenant_id]['refresh_token'] = tokens.refresh_token
 
     def set_refresh_token(self, token):
         """
@@ -849,6 +872,85 @@ class Operation(object):
         self.query_parameters = [p for _, p in op_desc.parameters.items() if p.location == ParameterLocation.QUERY]
         self.request_body = op_desc.request_body
 
+    def determine_tenant_id_for_service_request(self, **kwargs):
+        """
+        Determines the tenant_id for a service request from the kwargs to said
+        request and the application context (flask thread-local), if available.
+        """
+        # we need to determine the tenant id for the request.
+        tenant_id = None
+        # to begin, look for it in the parameters. if the caller explicitly set
+        # it, always use that.
+        try:
+            tenant_id = kwargs.get('_tapis_tenant_id')
+        except:
+            pass
+        # if the caller did not explicitly set the _tapis_tenant_id variable,
+        # we can look on the flask thread local for data about the request.
+        if not tenant_id:
+            try:
+                from flask import g
+                # g.x_tapis_tenant corresponds to the X-Tapis-Tenant OBO header. if this
+                # is set, it is always the tenant_id of the request.
+                tenant_id = getattr(g, 'x_tapis_tenant', None)
+                # g.request_tenant_id, if set, incorporates the tenant_id claim from
+                # the token and the OBO headers.
+                if not tenant_id:
+                    tenant_id = getattr(g, 'request_tenant_id', None)
+                # finally, the g.tenant_id represents the tenant_id claim in the access token.
+                # we should only use this if nothing else is set.
+                if not tenant_id:
+                    tenant_id = getattr(g, 'tenant_id', None)
+            # note that flask will throw a RuntimeError if attempting to access the thread local object, j, outside
+            # of an application context.
+            except (ImportError, RuntimeError):
+                pass
+        # check the headers to see if X-Tapis-Tenant is being set; if so, use that:
+        if not tenant_id and 'headers' in kwargs:
+            try:
+                tenant_id = kwargs['headers']['X-Tapis-Tenant']
+            except:
+                pass
+        # finally, look for a tenant_id on the tapis client itself
+        if not tenant_id:
+            tenant_id = self.tapis_client.tenant_id
+        return tenant_id
+
+    def determine_user_for_service_request(self, **kwargs):
+        """
+        Determines the username for a service request from the kwargs to said
+        request and the application context (flask thread-local), if available.
+        """
+        # we need to determine the user for the request.
+        user = None
+        # to begin, look for it in the parameters. if the caller explicitly set
+        # it, always use that.
+        try:
+            user = kwargs.get('_tapis_user')
+        except:
+            pass
+        # if the caller did not explicitly set the _tapis_user variable,
+        # we can look on the flask thread local for data about the request.
+        if not user:
+            try:
+                from flask import g
+                # g.x_tapis_user corresponds to the X-Tapis-User OBO header. if this
+                # is set, it is always the user of the request.
+                user = getattr(g, 'x_tapis_user', None)
+
+                # the g.username corresponds to the tapis/username claim in the access token.
+                # we should only use this if x_tapis_user was not set.
+                if not user:
+                    user = getattr(g, 'username', None)
+            # note that flask will throw a RuntimeError if attempting to access attributes on the thread local object,
+            # g, outside of an application context.
+            except (ImportError, RuntimeError):
+                pass
+        # finally, look a username on the tapis client itself
+        if not user:
+            user = self.tapis_client.username
+        return user
+
     def __call__(self, **kwargs):
         """
         Turns the operation object into a callable. Arguments must be passed as kwargs, where the name of each kwarg
@@ -862,12 +964,32 @@ class Operation(object):
         # the http method is defined by the operation -
         http_method = self.http_method.upper()
 
-        # construct the http path -
-        # some API definitions, such as SK, chose to not include the "/v3/" at the beginning of their paths, so we add it in:
-        if not self.op_desc.path_name.startswith('/v3/'):
-            self.url = f'{self.tapis_client.base_url}/v3{self.op_desc.path_name}'  # base url
+        # determine base_url --
+        base_url = None
+        # if this call is coming from another tapis service,
+        # we must determine base_url for a service to service request.
+        if self.tapis_client.is_tapis_service:
+            tenant_id = self.determine_tenant_id_for_service_request(**kwargs)
+            # if we got a tenant_id, use it to look up the base_url:
+            if tenant_id:
+                try:
+                    base_url = self.tapis_client.tenants.get_base_url_for_service_request(tenant_id, self.resource_name)
+                except:
+                    pass
+            # if all else fails, use the base_url for the client
+            if not base_url:
+                base_url = self.tapis_client.base_url
+
+        # otherwise, just use the base_url passed into the tapis client --
         else:
-            self.url = f'{self.tapis_client.base_url}{self.op_desc.path_name}'  # base url
+            base_url = self.tapis_client.base_url
+        # construct the http path -
+        # some API definitions, such as SK, chose to not include the "/v3/" at the beginning of their paths,
+        # so we add it in:
+        if not self.op_desc.path_name.startswith('/v3/'):
+            self.url = f'{base_url}/v3{self.op_desc.path_name}'  # base url
+        else:
+            self.url = f'{base_url}{self.op_desc.path_name}'  # base url
         url = self.url
         for param in self.path_parameters:
             # look for the name in the kwargs
@@ -903,14 +1025,28 @@ class Operation(object):
 
         # construct the http headers -
         headers = {}
-
+        access_token = None
         # set the X-Tapis-Token header using the client
-        if self.tapis_client.get_access_jwt():
+        if self.tapis_client.is_tapis_service and hasattr(self.tapis_client, 'service_tokens'):
+            # service_tokens may be defined but still be empty dictionaries... this __call__ could be to get
+            # the service's first set of tokens.
+            if tenant_id in self.tapis_client.service_tokens.keys() \
+                    and 'access_token' in self.tapis_client.service_tokens[tenant_id].keys():
+                try:
+                    access_token = self.tapis_client.service_tokens[tenant_id]['access_token']
+                except KeyError:
+                    raise errors.BaseTapyException(f"Did not find service tokens for tenant {tenant_id}; "
+                                                   f"service_tokens for tenant: {self.tapis_client.service_tokens[tenant_id]}")
+            else:
+                pass
+        elif self.tapis_client.get_access_jwt():
+            access_token = self.tapis_client.get_access_jwt()
+        if access_token:
             # check for a token about to expire in the next 5 seconds; assume by default we have a token with
             # plenty of time remaining.
             time_remaining = datetime.timedelta(days=10)
             try:
-                time_remaining = self.tapis_client.access_token.expires_in()
+                time_remaining = access_token.expires_in()
             except:
                 # it is possible the access_token does not have an expires_in attribute and/or that it is not
                 # callable. we just pass on these exceptions and do not try to refresh the token.
@@ -927,12 +1063,24 @@ class Operation(object):
                         # for now, if we get an error trying to refresh the tokens,s, we ignore it and try the
                         # request anyway.
                         pass
-            headers = {'X-Tapis-Token': self.tapis_client.get_access_jwt(), }
+            if self.tapis_client.is_tapis_service:
+                try:
+                    jwt = self.tapis_client.service_tokens[tenant_id]['access_token'].access_token
+                except KeyError:
+                    raise errors.BaseTapyException(f"No service tokens found for tenant {tenant_id} after refresh.")
+            else:
+                jwt = self.tapis_client.get_access_jwt()
+            headers = {'X-Tapis-Token':  jwt}
 
-        # the X-Tapis-Tenant and X-Tapis-Username headers can be set when the token represents a service account and the
-        # service is making a request on behalf of another user/tenant.
-        if self.tapis_client.x_tenant_id:
-            headers['X-Tapis-Tenant'] = self.tapis_client.x_tenant_id
+        # the X-Tapis-Tenant and X-Tapis-Username headers must be set when the token represents a service account.
+        # if this is a tapis service, we should have derived the correct tenant_id when we determined the tenant_id
+        # for the request, above.
+        if self.tapis_client.is_tapis_service:
+            if tenant_id:
+                headers['X-Tapis-Tenant'] = tenant_id
+        # similarly for username, we first look in the request content:
+        if self.tapis_client.is_tapis_service:
+            x_isername = self.determine_user_for_service_request(**kwargs)
         if self.tapis_client.x_username:
             headers['X-Tapis-User'] = self.tapis_client.x_username
 
@@ -958,8 +1106,8 @@ class Operation(object):
                     data = kwargs['request_body']
                 else:
                     # otherwise, the request body has defined properties, so look for each one in the function kwargs
-                    for p_name, p_desc in self.op_desc.request_body.content[
-                        'application/json'].schema.properties.items():
+                    for p_name, p_desc in \
+                            self.op_desc.request_body.content['application/json'].schema.properties.items():
                         if p_name in kwargs:
                             data[p_name] = kwargs[p_name]
                         elif p_name in required_fields:
@@ -1026,16 +1174,7 @@ class Operation(object):
         # resp.headers is a case-insensitive dict
         resp_content_type = resp.headers.get('content-type')
         if hasattr(resp_content_type, 'lower') and resp_content_type.lower() == 'application/json':
-            try:
-                json_content = resp.json()
-            except Exception as e:
-                msg = f'Requests could not produce JSON from the response even though the content-type was ' \
-                      f'application/json. Exception: {e}'
-                # TODO -- should this not be an error if the API has described the content-type as application/json?
-                #         what valid use cases do we still have for passing raw content in this case?
-                if debug:
-                    return resp.content, debug_data
-                return resp.content
+            json_content = resp.json()
             # get the Tapis result object which could be a JSON object or list.
             try:
                 result = json_content.get('result')
@@ -1043,7 +1182,7 @@ class Operation(object):
                 # some Tapis APIs, such as the Meta API, do not return "result" objects and/or do not return the
                 # standard Tapis stanzas but rather "raw" JSON documents
                 return resp.content
-            if result:
+            if result or result == [] or result == {}:
                 # if it is a list we should return a list of TapisResult objects:
                 if _seq_but_not_str(result):
                     if len([item for item in result if type(item) in TapisResult.PRIMITIVE_TYPES]) > 0:
@@ -1052,7 +1191,7 @@ class Operation(object):
                         return TapisResult(result)
                     else:
                         if debug:
-                            [TapisResult(**x) for x in result if not x == 'self'], debug_data
+                            return [TapisResult(**x) for x in result if not x == 'self'], debug_data
                         return [TapisResult(**x) for x in result if not x == 'self']
                 # otherwise, assume it is a JSON object and return that directly as a result -
                 try:
@@ -1135,12 +1274,19 @@ class TapisResult(object):
 
     def __str__(self):
         attrs = '\n'.join([f'{str(a)}: {getattr(self, a)}' for a in dir(self) if
-                           not a.startswith('__') and not a.startswith('PRIMITIVE_TYPES')])
+                           not a == 'get' and not a.startswith('__') and not a.startswith('PRIMITIVE_TYPES')])
         return f'\n{attrs}'
 
     def __repr__(self):
         return str(self)
 
+    def get(self, attr, default=None):
+        """Provide a dictionary-like get() syntax """
+        try:
+            return getattr(self, attr)
+        except:
+            return default
+
 
 class Debug(object):
     """
